/* -------------------------------------------------------
                        tastdemo.c

    Versuche mit einer 4x4 Tastaturmatrix

     MCU      :  STM8S103F3

     Hardware :  PCF8574AP
                 4x4 Tastaturmatrix
                 LED
     Takt     :  interner Takt 16 MHz

     29.11.2017  R. Seelig
   ------------------------------------------------------ */

/*

            ################################
            STM8S103F3P6 Pinout
            ################################

                            ------------
UART1_CK / TIM2_CH1 / PD4  |  1     20  |  PD3 / AIN4 / TIM2_CH2 / ADC_ETR
    UART1_TX / AIN5 / PD5  |  2     19  |  PD2 / AIN3
    UART1_RX / AIN6 / PD6  |  3     18  |  PD1 / SWIM
                     NRST  |  4     17  |  PC7 / SPI_MISO
              OSCIN / PA1  |  5     16  |  PC6 / SPI_MOSI
             OSCOUT / PA2  |  6     15  |  PC5 / SPI_CLK
                Vss (GND)  |  7     14  |  PC4 / TIM1_CH4 / CLK_CCO / AIN2
                VCAP (*1)  |  8     13  |  PC3 / TIM1_CH3 /
                Vdd (+Ub)  |  9     12  |  PB4 / I2C_SCL
           TIM2_CH3 / PA3  | 10     11  |  PB5 / I2C_SDA
                            ------------


            ################################
            PCF8574A I2C I/O Expander Pinout
            ################################

                            ---------
                        A0 | 1    16 | Vcc
                        A1 | 2    15 | SDA
                        A2 | 3    14 | SCL
                        P0 | 4    13 | /INT
                        P1 | 5    12 | P7
                        P2 | 6    11 | P6
                        P3 | 7    10 | P5
                       GND | 8     9 | P4
                            ---------


             ##############################
             I2C Anbindung an STM8S103F3P6:
             ##############################

             SDA : PB5 (Pin 11)
             SCL : PB4 (Pin 12)


             ####################################
             Anschluss Tastaturmatrix an PC8574AP
             ####################################


                    C1         C2         C3         C4
                   (P4)       (P5)       (P6)       (P7)

                    o          o          o          o
                    |          |          |          |
                    |          |          |          |
                S1  |      S2  |      S3  |      S4  |
             o-- \--o   o-- \--o   o-- \--o   o-- \--o
             |      |   |      |   |      |   |      |
    R1 o-----o----------o----------o----------o      |
   (P0)             |          |          |          |
                    |          |          |          |
                S5  |      S6  |      S7  |      S8  |
             o-- \--o   o-- \--o   o-- \--o   o-- \--o
             |      |   |      |   |      |   |      |
    R2 o-----o----------o----------o----------o      |
   (P1)             |          |          |          |
                    |          |          |          |
                S9  |      S10 |      S11 |      S12 |
             o-- \--o   o-- \--o   o-- \--o   o-- \--o
             |      |   |      |   |      |   |      |
    R3 o-----o----------o----------o----------o      |
   (P2)             |          |          |          |
                    |          |          |          |
                S13 |      S14 |      S15 |      S16 |
             o-- \--o   o-- \--o   o-- \--o   o-- \--o
             |      |   |      |   |      |   |
    R4 o-----o----------o----------o----------o
   (P3)


             ####################################
             IDC-10 Connector Tastaturmatrix
             ####################################


                       R2  R4  C2  C4  nc
                      -------------------
                   2 | o   o   o   o   o | 10
                   1 | o   o   o   o   o | 9
                      --------###--------
                       R1  R3  C1  C3  nc


             ##############################
             Tastaturlayout
             ##############################


          +-------+  +-------+  +-------+  +-------+
          |   1   |  |   2   |  |   3   |  |  ESC  |
          |   @   |  |  abc  |  |  def  |  |       |
          +-------+  +-------+  +-------+  +-------+
          +-------+  +-------+  +-------+  +-------+
          |   4   |  |   5   |  |   6   |  |  DEL  |
          |  ghi  |  |  jkl  |  |  mno  |  |       |
          +-------+  +-------+  +-------+  +-------+
          +-------+  +-------+  +-------+  +-------+
          |   7   |  |   8   |  |   9   |  | SHIFT |
          | pqrs  |  |  tuv  |  | wxyz  |  | LOCK  |
          +-------+  +-------+  +-------+  +-------+
          +-------+  +-------+  +-------+  +-------+
          |   .   |  |   0   |  |   =   |  |  RET  |
          | * /+- |  | SPACE |  | (){}  |  |       |
          +-------+  +-------+  +-------+  +-------+

   Shift-Lock Leuchtdiode an Kathode an PD3, Anode an Vcc


*/

#include "stm8s.h"
#include "stm8_init.h"
#include "stm8_gpio.h"
#include "my_printf.h"
#include "uart.h"
#include "sw_i2c.h"


#define shiftled_init()     PD3_output_init()
#define shiftled_clr()      PD3_set()
#define shiftled_set()      PD3_clr()

#define printf              my_printf

#define pcfaddr             0x40                  // I2C Adresse PCF8574 I/O Expander


/* ----------------------------------------------
     globale Variable
   ---------------------------------------------- */
uint8_t shiftlock= 0;                             // 0: Rueckgabe als Kleinbuchstaben
                                                  // 1: Rueckgabe als Grossbuchstaben
uint8_t keypresent= 0;                            // 0: keine Taste gedrueckt, 1: Taste wurde gedrueckt
uint8_t keybuf;                                   // zuletzt gedrueckte Taste
uint8_t keyticks;                                 // Anzahl der Hits, wie oft eine Taste gedrueckt wurde

/* ----------------------------------------------
     Tastaturlayout (mit Mehrfachbelegung)
   ---------------------------------------------- */
uint8_t keylayout[16][5] =
{
  { '1', '@', 0,  0,  0 },
  { '2', 'A','B','C', 0 },
  { '3', 'D','E','F', 0 },

  { 27 ,  0,  0,  0,  0 },

  { '4', 'G','H','I', 0 },
  { '5', 'J','K','L', 0 },
  { '6', 'M','N','O', 0 },

  { 8 ,  0,  0,  0,  0 },

  { '7', 'P','Q','R','S'},
  { '8', 'T','U','V', 0 },
  { '9', 'W','X','Y','Z'},

  { 1  ,  0,  0,  0,  0 },

  { '.', '*','/','+','-'},
  { '0', ' ', 0,  0,  0 },
  { '=', '(',')','{','}'},

  { 13 ,  0,  0,  0,  0 }

};


void putchar(char ch);


char mylog2(uint8_t value)
{
  uint8_t i;

  if (value== 1) return 0;
  for (i= 1; i< 8; i++)
    if ((1 << i) >= value) return i;
  return 255;
}

void t4x4_gettastvalues(uint8_t *values)
{
  uint8_t i;

  for (i= 0; i< 4; i++)
  {
    i2c_sendstart();
    i2c_write(pcfaddr);                      // Adresse PCF8574
    i2c_write(~(1 << i));
    i2c_stop();
    i2c_sendstart();
    i2c_write(0x41);
    i2c_read_ack();
    delay_us(10);
    *values= (~(i2c_read_nack() >> 4)) & 0x0f;
    i2c_stop();
    values++;
  }

}

/* ------------------------------------------------------
                     tim2_init
      Initialisiert Timer2 als Overflow-Interrupt-Timer
      (Interruptvector 13)

      Berechnung:

      tim2_takt  = F_CPU / 2 ^ scalvalue
      time       = tim2_takt * (relvalue -1)

      Bsp.:
      scalvalue = 0;         Teilertakt / 1 => 0.0625 uS
      relvalue  = 15999;     time= (relvalue+1) * scalvalue = 16000 * 0.0625 uS
                                                            = 1 mS


      retvalue  = 15625
      scalvalue = 9
      =>          0.5 s

      retvalue = 319
      scalvalue = 9
      =>          10 ms


   ------------------------------------------------------ */
void tim2_init(void)
{

  // Aufruf Interrupt alle 5 mS
  #define relvalue    159
  #define scalvalue   9

  TIM2_PSCR= scalvalue;         // F_CPU / 2 ^  scalvalue
  TIM2_IER |= TIM2_IER_UIE;     // enable overflow
  TIM2_CR1 |= TIM2_CR1_CEN;     // Timer2 enable ( CEN = ClockENable )
  TIM2_CR1 |= TIM2_CR1_ARPE;    // Timer2 autoreload ( mit Werten in TIM2_ARR )

  #if (relvalue > 255)
    TIM2_ARRH= (uint8_t) (relvalue >> 8);
  #else
    TIM2_ARRH= 0;
  #endif
  TIM2_ARRL= (uint8_t) (relvalue & 0x00ff);

  int_enable();
}


void tim2_ovf(void) __interrupt(13)
// 13 ist der Interruptvektor fuer Timer2 overflow
{

  volatile uint8_t  tastvalues[4] = { 0,0,0,0 };
  uint8_t           i;
  static uint8_t    taste = 0;
  static uint8_t    oldtaste = 0;
  static uint8_t    tastcnt= 0;
  static uint8_t    intcnt= 0;
  static uint8_t    pressed = 0;

  if (intcnt> 8)
  {
    if (intcnt== 65)                                // es kam innerhalb 65 Interruptaufrufen keine weiteren
                                                    // Tastenklicks
    {
      if (oldtaste)
      {
        keybuf= oldtaste;
        keyticks= tastcnt;                          // beinhaltet, wie oft ein und dieselbe Taste gedrueckt wurde
        keypresent= 1;                              // Anzeigeflag, dass eine Taste gedrueckt wurde
      }
      tastcnt= 0;
      intcnt= 0;
      taste= 0; oldtaste= 0;
    }

    t4x4_gettastvalues(&tastvalues[0]);            // Tastaturmatrix ueber I2C scannen

    taste= 0; i= 0;
    do
    {
      if ((tastvalues[i]> 0))                      // wenn Wert != 0 wurde eine Taste gedrueckt
      {
        taste= (i*4) + mylog2(tastvalues[i]) + 1;
      }
      i++;
    }while ((i< 4) && (!(taste)));                 // alle 4 Bitpositionen scannen

    if (taste== oldtaste)                          // wenn es die gleiche Taste war...
    {
      if (!(pressed))
      {
        tastcnt++;                                 // ... gedrueckte Tastenanzahl erhoehen
        intcnt= 0;
      }
    }
    else
    {
      if (taste)
      {
        intcnt= 0;
        oldtaste= taste;
        tastcnt= 1;
      }
    }
    if (taste) pressed= 1; else pressed= 0;
  }
  intcnt++;

  TIM2_SR1 &= ~TIM2_SR1_UIF;        // Interrupt quittieren
}


char t4x4_ischar(void)
{
  return keypresent;
}

char t4x4_getchar(void)
{

  keypresent= 0;
  return keylayout[keybuf-1][keyticks-1];

}

void show_keylayout(void)
{
  printf("\n\r");
  printf("\n\rTastaturlayout:");
  printf("\n\r");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r|   1   |  |   2   |  |   3   |  |  ESC  |");
  printf("\n\r|   @   |  |  abc  |  |  def  |  |       |");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r|   4   |  |   5   |  |   6   |  |  DEL  |");
  printf("\n\r|  ghi  |  |  jkl  |  |  mno  |  |       |");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r|   7   |  |   8   |  |   9   |  | SHIFT |");
  printf("\n\r| pqrs  |  |  tuv  |  | wxyz  |  | LOCK  |");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r|   .   |  |   0   |  |   =   |  |  RET  |");
  printf("\n\r| * /+- |  | SPACE |  | (){}  |  |       |");
  printf("\n\r+-------+  +-------+  +-------+  +-------+");
  printf("\n\r");
  printf("\n\r");
}



/* ---------------------------------------------------------------------
                              MAIN
   --------------------------------------------------------------------- */

int main(void)
{
  uint8_t ch;

  sysclock_init(0);

  i2c_master_init();
  shiftled_init();
  shiftled_clr();

  uart_init(19200);
  tim2_init();

  printf("\n\n\rTastaturmatrix 4x4 mit STM8S103F3P6");
  printf("\n\rund Software I2c");
  printf("\n\r-----------------------------------\n\n\r");

  show_keylayout();

  printf("\n\r");

  do
  {
    if (t4x4_ischar())
    {
      ch= t4x4_getchar();
      if (ch>= ' ')
      {
        if (!shiftlock)
        {
          if ((ch>= 'A') && (ch<= 'Z')) ch += 32;
        }
        printf("%c",ch);
      }
      switch (ch)
      {
        case 0x0d : printf("\n\r"); break;
        case 0x01 : {
                      if (ch== 1)
                      {
                        if (shiftlock)
                        {
                          shiftlock= 0;
                          shiftled_clr();
                        }
                        else
                        {
                          shiftlock= 1;
                          shiftled_set();
                        }
                      }
                      break;
                    }
        case 0x08 : printf("%c",8);
        default : break;
      }
      delay(50);
    }
    else ch= 0;
  } while(ch!= 27);
  printf("\n\n\rProgram stopped\n\r");
  while(1);
}


// #################################################################################################

/* --------------------------------------------------
     PUTCHAR

     Diese Funktion wird von < my_printf > zur Ausgabe
     von Zeichen aufgerufen. Hier ist eine Hardware-
     zeichenausgabefunktion anzugeben, mit der
     my_printf zusammen arbeiten soll.
   -------------------------------------------------- */
void putchar(char ch)
{
  uart_putchar(ch);
}

